from rasterio import open as rasterio_open
from rasterio.mask import mask as rasterio_mask
from os import getcwd, makedirs
from os.path import isfile, join, exists
from numpy import stack, float32
from skimage.transform import resize
from skimage.color import rgb2hsv, hsv2rgb
from shapely import from_wkt
from shapely.geometry import Polygon
from typing import Dict, Union
from geomet import wkt


def rgbn_composite(
    red: str,
    green: str,
    blue: str,
    nir: str = None,
    outdir: str = getcwd(),
    filename: str = "rgbn_composite.tif",
):
    """
    Stack bands

    Args:
        red: Red channel
        green: Green channel
        blue: Blue channel
        nir: (Optional) Nir channel
        outdir: Output path
        filename: Output filename
    Returns:
        GeoTIFF file
    """
    if isfile(red) and isfile(green) and isfile(blue):
        bands = [
            rasterio_open(red),
            rasterio_open(green),
            rasterio_open(blue),
        ]

        if nir is not None and isfile(nir):
            bands.append(rasterio_open(nir))
        elif nir is not None and not isfile(nir):
            raise FileNotFoundError("Check band's file path")

        if not exists(outdir):
            makedirs(outdir)

        bands_metadata = bands[0].meta.copy()

        merged = stack([band.read(1) for band in bands])

        for band in bands:
            band.close()

        count, height, width = merged.shape

        bands_metadata.update(width=width, height=height, count=count, nodata=0)

        with rasterio_open(join(outdir, filename), "w", **bands_metadata) as raster:
            raster.write(merged)

    else:
        raise FileNotFoundError("Check band's file path")


def pansharpening(
    panchromatic: str = None,
    multispectral: str = None,
    outdir: str = getcwd(),
    filename: str = "pansharp.tif",
):
    """
    Pansharpen multispectral file

    Args:
        panchromatic: Panchromatic band (Band 0)
        multispectral: Multispectral band
        outdir: Output Directory
        filename: Output file name
    Returns:
        GeoTIFF file
    """

    if isfile(panchromatic) and isfile(multispectral):
        if not exists(outdir):
            makedirs(outdir)

        BIT_DEPTH = 65535

        with rasterio_open(multispectral) as multispectral_file:
            multispectral_array = multispectral_file.read().astype(
                dtype=float32, copy=False
            )

        # Normalize RGB to 0..1 interval
        multispectral_array = multispectral_array / BIT_DEPTH

        multispectral_hsv = rgb2hsv(multispectral_array, channel_axis=0)

        del multispectral_array

        with rasterio_open(panchromatic) as panchromatic_file:
            # Create metadata copy and add expected output data
            panchromatic_metadata = panchromatic_file.meta.copy()
            panchromatic_metadata.update(count=3, dtype="float32")

            # Get matrix data
            panchromatic_array = panchromatic_file.read(1).astype(
                dtype=float32, copy=False
            )

        height, width = panchromatic_array.shape

        # Resizing all bands to pansharp dimensions
        multispectral_hsv = resize(
            multispectral_hsv, (3, height, width), anti_aliasing=False
        )

        # normalize pan to 0..1 interval
        panchromatic_array = panchromatic_array / BIT_DEPTH

        # Replacing Value component by Panchromatic
        multispectral_hsv[2, :, :] = panchromatic_array

        del panchromatic_array

        with rasterio_open(
            join(outdir, filename), "w", **panchromatic_metadata
        ) as raster:
            raster.write(hsv2rgb(multispectral_hsv, channel_axis=0))

    else:
        raise FileNotFoundError("Invalid files")


def clip(
    raster: str,
    mask: Union[Dict, Polygon],
    outdir: str = getcwd(),
    filename: str = "raster_clip.tif",
    **kwargs,
):
    """
    Clip raster

    Args:
        raster: Image to clip
        mask: Area to use as clip mask
        outdir: Output Directory
        filename: Output file name
        kwargs: Any option you want to add in rasterio mask method
    Returns:
        GeoTIFF file
    """
    if isfile(raster):
        if isinstance(mask, Dict):
            if mask.get("type") in ["Feature", "FeatureCollection"]:
                # Converting to Shapely Polygon to assure crop method will recognize the coordinates
                if mask.get("type") == "Feature":
                    mask = from_wkt(wkt.dumps(mask["geometry"], decimals=10))
                elif mask.get("type") == "FeatureCollection":
                    mask = from_wkt(
                        wkt.dumps(mask["features"][0]["geometry"], decimals=10)
                    )
            else:
                raise ValueError("Mask Invalid")
        elif isinstance(mask, Polygon):
            if not mask.is_valid:
                raise ValueError("Mask Invalid")

        if not exists(outdir):
            makedirs(outdir)

        with rasterio_open(raster) as raster_file:
            raster_metadata = raster_file.meta.copy()

            masked, transform = rasterio_mask(
                dataset=raster_file, shapes=[mask], crop=True, **kwargs
            )

            count, height, width = masked.shape

            raster_metadata.update(transform=transform, height=height, width=width)

            with rasterio_open(join(outdir, filename), "w", **raster_metadata) as dst:
                dst.write(masked)

    else:
        raise FileNotFoundError("Invalid Raster File")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from concurrent.futures import ThreadPoolExecutor
from .cbers4a import Search, ItemCollection
from datetime import date
from typing import List, overload, Dict, Union, Tuple
from os.path import isdir, join
from os import getcwd, cpu_count, mkdir
from geopandas import GeoDataFrame
from pandas import json_normalize
from shapely.geometry import Polygon


class Cbers4aAPI:
    def __init__(self, user):
        self._user = user

    @property
    def user(self):
        return self._user

    @user.setter
    def user(self, user):
        self._user = user

    @staticmethod
    def query(
        *,
        location: Union[List[float], Tuple, Polygon],
        initial_date: date,
        end_date: date,
        cloud: int,
        limit: int,
        collections: List[str] = None,
    ):
        """
        Query Images from INPE's catalog

        Args:
            location: Bounding box or Path Row
            initial_date: Images from this date
            end_date: Images to this date
            cloud: Percentage of clouds
            limit: Limit of returned images
            collections: Collection's name(s)

        Notes:
            Location:
                - Bounding box: `location=[-0.5, 1.0, 0.5, -0.5]`
                - Path and row respectively: `location=(225,75)`
            Initial and end dates:
                - Use date object from datetime library: `datetime.date(2021,05,25)`
                - Caution to end date always be bigger than initial date
                - e.g. `initial_date=datetime.date(2021,08,15)`
            Cloud coverage:
                - Search for scenes below certain cloud coverage percentage
                - e.g. `cloud=70`
            Limit:
                - Max value of returned features is 1000
                - e.g. `limit=500`
            Collections:
                - Available collections from INPE catalog <http://www2.dgi.inpe.br/catalogo/explore>
                - Always an array of collections name.
                - e.g. `['AMAZONIA1_WFI_L2_DN']` or `['AMAZONIA1_WFI_L2_DN', 'CBERS4A_WPM_L4_DN']` etc.

        Returns:
            dict: Dict with GeoJSON-like format

        """
        search = Search()

        if not location:
            raise Exception("Location cannot be empty")

        if isinstance(location, list):
            search.bbox(location)
        elif isinstance(location, tuple):
            path, row = location
            search.path_row(path, row)
        elif isinstance(location, Polygon):
            search.bbox(list(location.bounds))
        else:
            raise Exception("Provide a bbox or a path row")

        search.date(initial_date.isoformat(), end_date.isoformat())
        search.cloud_cover("<=", cloud)
        search.limit(limit)

        if collections is not None:
            search.collections(collections)

        result = search()
        return result.featurescollection

    @staticmethod
    def query_by_id(id_: Union[str, List[str]]):
        """
        Search a product by id

        Args:
            id_: One or more scene's id
        Returns:
            dict: Dict with GeoJSON-like format
        """
        search = Search()

        # search() method only works with a list of ids.
        if type(id_) is not list:
            search.ids([id_])
        else:
            search.ids(id_)

        result = search()
        return result.featurescollection

    # AUXILIARY -------Description------------------
    #
    # _check_for_exception -> Private method specialized to check for file and folders and mistakes
    # ----------------------------------------------

    def __check_for_exception(self, product, bands, outdir):
        """Search for exceptions block: empty product, empty bands, output dir does not exist"""

        # Check for empty products in different contexts
        if isinstance(product, Dict):
            if not product:  # Check if dictionary is empty
                raise TypeError("No product to download")
        elif isinstance(product, GeoDataFrame):
            if product.empty:  # Check if data frame is empty
                raise TypeError("No product to download")
        # ---
        if not bands or bands == [""]:
            raise TypeError("Choose bands to download")
        elif not isdir(outdir):
            raise NotADirectoryError("Choose a valid output directory")

    def __download(
        self,
        products: Dict,
        bands: List[str],
        threads: int = cpu_count(),
        outdir: str = getcwd(),
        with_folder: bool = False,
    ):
        self.__check_for_exception(products, bands, outdir)

        products = ItemCollection(products).items()

        with ThreadPoolExecutor(
            max_workers=threads, thread_name_prefix="cbers4a"
        ) as pool:
            if with_folder:
                root = outdir  # Save the outdir start point (to looks like a pushd and popd)
            for product in products:
                if with_folder:
                    new_path = join(root, product.id)
                    mkdir(new_path)
                    outdir = new_path
                for band in bands:
                    pool.submit(product.download, band, self._user, outdir)

    def __download_gdf(
        self,
        products: GeoDataFrame,
        bands: List[str],
        threads: int = cpu_count(),
        outdir: str = getcwd(),
        with_folder: bool = False,
    ):
        self.__check_for_exception(products, bands, outdir)

        with ThreadPoolExecutor(
            max_workers=threads, thread_name_prefix="cbers4a_gdf"
        ) as pool:
            if with_folder:
                root = outdir  # Save the outdir start point (to looks like a pushd and popd)
            for index, row in products.iterrows():
                if with_folder:
                    new_path = join(root, str(index))
                    mkdir(new_path)
                    outdir = new_path

                products_query = self.query_by_id(str(index))
                products_query = ItemCollection(products_query).items()
                for product in products_query:
                    for band in bands:
                        pool.submit(product.download, band, self._user, outdir)

    @overload
    def download(
        self,
        products: Dict,
        bands: List[str],
        threads: int = cpu_count(),
        outdir: str = getcwd(),
        with_folder: bool = False,
    ):
        ...

    @overload
    def download(
        self,
        products: GeoDataFrame,
        bands: List[str],
        threads: int = cpu_count(),
        outdir: str = getcwd(),
        with_folder: bool = False,
    ):
        ...

    def download(
        self,
        products,
        bands,
        threads=cpu_count(),
        outdir=getcwd(),
        with_folder: bool = False,
    ):
        """
        Download bands from all given scenes

        Args:
            products: Data returned from API
            bands: List of band color's name
            threads: Max of cpu threads
            outdir: Output path
            with_folder: Group scene bands in a sub folder
        Examples:
            - download(my_query_result, ['red', 'green'], 3, './downlaods', true)
            - download(my_query_result, ['red'], outdir='./downloads', with_folder=true)
        Returns:
            GeoTIFF files
        """
        if isinstance(products, dict):
            return self.__download(products, bands, threads, outdir, with_folder)
        elif isinstance(products, GeoDataFrame):
            return self.__download_gdf(products, bands, threads, outdir, with_folder)
        else:
            raise TypeError("Bad Arguments")

    @staticmethod
    def to_geodataframe(products: Dict, crs: str = "EPSG:4326"):
        """
        Transform products list to a GeoDataFrame

        Args:
            products: GeoJSON-like dictionary returned from API
            crs: Coordinate Reference System (ex: EPSG:4326)
        Returns:
            GeoDataFrame
        """
        return GeoDataFrame.from_features(products, crs=crs).set_index(
            json_normalize(products["features"])["id"].values
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from datetime import datetime
import requests
from os.path import join


class Item(object):
    """Class to parse items from INPE STAC Catalog"""

    def __init__(self, feature):
        self._feature = feature

    def __str__(self):
        """
        docstring
        """
        row = "{id:<25s} {date:^10s} {cloud:>15.1f}".format
        return row(id=self.id, date=self.date, cloud=self.get_property("cloud_cover"))

    def _repr_html_(self):
        """
        html representation for jupyter notebook
        """

        html = f"""
        <html>
        <body>
        <img
            style="display: block; margin-left: auto; margin-right: auto;"
            src="{self.thumbnail}" width="296" border="0"
        />
        </body>
        </html>
        """

        return html

    def repr_html(self, width=450):
        """
        html representation to generate a html catalog
        """
        html = f"""
        <div>
            <h2>{self.id}</h2>
            <span>Satelitte: {self.satellite}</span>
            <span>Sensor: {self.sensor}</span>
            <span>Collection: {self.collection}</span>
            <span>Cloud cover: {self.cloud_cover}</span>
            <span>Datetime: {self.get_datetime_fmt(fmt='%Y-%m-%d-%H:%m:%S')}</span>
            <span>bbox: {self.bbox}</span>
            <img src="{self.thumbnail}" width="{width}"/>
        </div>
        """
        return html

    def __lt__(self, other):
        """
        docstring
        """
        return self.get_property("cloud_cover", 0) < other.get_property(
            "cloud_cover", 0
        )

    @property
    def __geo_interface__(self):
        """
        Simple GeoJSON-like interface
        """
        return self._feature["geometry"]

    @property
    def id(self):
        """
        Provider identifier globally unique.
        """
        return self._feature["id"]

    @property
    def date(self):
        """
        The acquisition date of the asset.
        """
        return self.get_datetime_fmt()

    @property
    def bbox(self):
        """
        Bounding Box of the asset.
        """
        return self._feature.get("bbox")

    @property
    def collection(self):
        """
        The id of the STAC Collection this Item references to.
        """
        return self._feature.get("collection")

    @property
    def thumbnail(self):
        """
        Thumbnail image link.
        """
        return self._feature["assets"]["thumbnail"]["href"]

    @property
    def assets(self):
        """
        Unique keys of asset objects that can be downloaded.
        """
        return list(self._feature["assets"].keys())

    @property
    def cloud_cover(self):
        """
        Cloud coverage %
        """
        return self.get_property("cloud_cover")

    @property
    def path_row(self):
        """
        Return a string 'PATH_ROW'
        """
        return (
            str(self.get_property("path", "")) + "_" + str(self.get_property("row", ""))
        )

    @property
    def sensor(self):
        """
        Return the sensor
        """
        return self.get_property("sensor")

    @property
    def satellite(self):
        """
        Return the satellite
        """
        return self.get_property("satellite")

    @property
    def html(self):
        """
        Html to show image preview
        """

        return self.repr_html()

    def get_datetime(self):
        """
        The acquisition date and time of the assets
        as an object of the class datetime.datetime.
        """
        try:
            return datetime.strptime(self.get_property("datetime"), "%Y-%m-%dT%H:%M:%S")
        except (ValueError, TypeError):
            return None

    def get_datetime_fmt(self, fmt="%Y-%m-%d"):
        """
        Return a string representing the acquisition date and time, controlled by a format string.
        """
        try:
            return self.get_datetime().strftime(fmt)
        except AttributeError:
            return None

    def get_property(self, p, v=None):
        """
        Get additional metadata fields.
        None if it does not exist.
        """
        return self._feature["properties"].get(p, v)

    def url(self, asset):
        """
        Get asset url
        """
        return self._feature["assets"][asset]["href"]

    def download(self, asset, credential, outdir, session=requests.Session()):
        """
        Download the asset.
        """
        url = self.url(asset)
        filename = url.split("/")[-1]
        outfile = join(outdir, filename)
        r = session.get(
            url,
            params={
                "email": credential,
                "item_id": self.id,
                "collection": self.collection,
            },
            stream=True,
            allow_redirects=True,
        )
        if r.status_code == 200:
            with open(outfile, "wb") as f:
                for ch in r.iter_content(chunk_size=4096):  # (page size 4Kb)
                    if ch:
                        f.write(ch)

            return outfile
        else:
            return "ERROR in " + url + " (" + r.reason + ")"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


class Collections(object):
    """Simple class to parse collections json from INPE STAC"""

    def __init__(self, data):
        """
        docstring
        """
        self._data = data

    def __getitem__(self, k):
        """
        docstring
        """
        try:
            return next(filter(lambda i: i["id"] == k, self._data["collections"]))
        except StopIteration:
            raise KeyError(k)

    def __iter__(self):
        """
        docstring
        """
        for i in [
            (collection["id"], collection["description"])
            for collection in self._data["collections"]
        ]:
            yield i

    def get_spatial_extent(self, id):
        """
        docstring
        """
        return self.__getitem__(id)["extent"]["spatial"]

    def get_temporal_extent(self, id):
        """
        docstring
        """
        return self.__getitem__(id)["extent"]["temporal"]
